@@ -25,6 +25,7 @@ THE SOFTWARE.
 """
 
 
+import re
 from functools import reduce, wraps
 import operator
 import sys
@@ -37,9 +38,6 @@ import builtins
 from sys import intern
 
 
-decorator_module = __import__("decorator", level=0)
-my_decorator = decorator_module.decorator
-
 # These are deprecated and will go away in 2022.
 all = builtins.all
 any = builtins.any
@@ -73,6 +71,7 @@ Memoization
 .. autofunction:: memoize_in
 .. autofunction:: keyed_memoize_on_first_arg
 .. autofunction:: keyed_memoize_method
+.. autofunction:: keyed_memoize_in
 
 Argmin/max
 ----------
@@ -93,9 +92,7 @@ Permutations, Tuples, Integer sequences
 .. autofunction:: wandering_element
 .. autofunction:: generate_nonnegative_integer_tuples_below
 .. autofunction:: generate_nonnegative_integer_tuples_summing_to_at_most
-.. autofunction:: generate_all_nonnegative_integer_tuples
 .. autofunction:: generate_all_integer_tuples_below
-.. autofunction:: generate_all_integer_tuples
 .. autofunction:: generate_permutations
 .. autofunction:: generate_unique_permutations
 
@@ -143,11 +140,6 @@ Helpers for :mod:`numpy`
 Timing data
 -----------
 
-.. data:: SUPPORTS_PROCESS_TIME
-
-   A :class:`bool` indicating whether :class:`ProcessTimer` measures elapsed
-   process time (available on Python 3.3+).
-
 .. autoclass:: ProcessTimer
 
 Log utilities
@@ -163,6 +155,16 @@ Sorting in natural order
 .. autofunction:: natorder
 .. autofunction:: natsorted
 
+Backports of newer Python functionality
+---------------------------------------
+
+.. autofunction:: resolve_name
+
+Hashing
+-------
+
+.. autofunction:: unordered_hash
+
 Type Variables Used
 -------------------
 
@@ -613,43 +615,48 @@ def memoize(*args: F, **kwargs: Any) -> F:
             % ", ".join(list(kwargs.keys())))
 
     if key_func is not None:
-        @my_decorator
-        def _deco(func, *args, **kwargs):
-            # by Michele Simionato
-            # http://www.phyast.pitt.edu/~micheles/python/
-            key = key_func(*args, **kwargs)
-            try:
-                return func._memoize_dic[key]  # pylint: disable=protected-access
-            except AttributeError:
-                # _memoize_dic doesn't exist yet.
-                result = func(*args, **kwargs)
-                func._memoize_dic = {key: result}  # pylint: disable=protected-access
-                return result
-            except KeyError:
-                result = func(*args, **kwargs)
-                func._memoize_dic[key] = result  # pylint: disable=protected-access
-                return result
+        def _decorator(func):
+            def wrapper(*args, **kwargs):
+                key = key_func(*args, **kwargs)
+                try:
+                    return func._memoize_dic[key]  # noqa: E501 # pylint: disable=protected-access
+                except AttributeError:
+                    # _memoize_dic doesn't exist yet.
+                    result = func(*args, **kwargs)
+                    func._memoize_dic = {key: result}  # noqa: E501 # pylint: disable=protected-access
+                    return result
+                except KeyError:
+                    result = func(*args, **kwargs)
+                    func._memoize_dic[key] = result  # noqa: E501 # pylint: disable=protected-access
+                    return result
+
+            from functools import update_wrapper
+            update_wrapper(wrapper, func)
+            return wrapper
+
     else:
-        @my_decorator
-        def _deco(func, *args):
-            # by Michele Simionato
-            # http://www.phyast.pitt.edu/~micheles/python/
-            try:
-                return func._memoize_dic[args]  # pylint: disable=protected-access
-            except AttributeError:
-                # _memoize_dic doesn't exist yet.
-                result = func(*args)
-                func._memoize_dic = {args: result}  # pylint:disable=protected-access
-                return result
-            except KeyError:
-                result = func(*args)
-                func._memoize_dic[args] = result  # pylint: disable=protected-access
-                return result
+        def _decorator(func):
+            def wrapper(*args):
+                try:
+                    return func._memoize_dic[args]  # noqa: E501 # pylint: disable=protected-access
+                except AttributeError:
+                    # _memoize_dic doesn't exist yet.
+                    result = func(*args)
+                    func._memoize_dic = {args: result}  # noqa: E501 # pylint:disable=protected-access
+                    return result
+                except KeyError:
+                    result = func(*args)
+                    func._memoize_dic[args] = result  # noqa: E501 # pylint: disable=protected-access
+                    return result
+
+            from functools import update_wrapper
+            update_wrapper(wrapper, func)
+            return wrapper
 
     if not args:
-        return _deco
+        return _decorator  # type: ignore
     if callable(args[0]) and len(args) == 1:
-        return _deco(args[0])
+        return _decorator(args[0])
     raise TypeError(
         "memoize received unexpected position arguments: %s" % args)
 
@@ -669,8 +676,9 @@ def memoize_on_first_arg(function, cache_dict_name=None):
     """
 
     if cache_dict_name is None:
-        cache_dict_name = intern("_memoize_dic_"
-                + function.__module__ + function.__name__)
+        cache_dict_name = intern(
+                f"_memoize_dic_{function.__module__}{function.__name__}"
+                )
 
     def wrapper(obj, *args, **kwargs):
         if kwargs:
@@ -681,16 +689,20 @@ def memoize_on_first_arg(function, cache_dict_name=None):
         try:
             return getattr(obj, cache_dict_name)[key]
         except AttributeError:
-            result = function(obj, *args, **kwargs)
-            setattr(obj, cache_dict_name, {key: result})
-            return result
+            attribute_error = True
         except KeyError:
-            result = function(obj, *args, **kwargs)
+            attribute_error = False
+
+        result = function(obj, *args, **kwargs)
+        if attribute_error:
+            object.__setattr__(obj, cache_dict_name, {key: result})
+            return result
+        else:
             getattr(obj, cache_dict_name)[key] = result
             return result
 
     def clear_cache(obj):
-        delattr(obj, cache_dict_name)
+        object.__delattr__(obj, cache_dict_name)
 
     from functools import update_wrapper
     new_wrapper = update_wrapper(wrapper, function)
@@ -701,9 +713,15 @@ def memoize_on_first_arg(function, cache_dict_name=None):
 
 def memoize_method(method: F) -> F:
     """Supports cache deletion via ``method_name.clear_cache(self)``.
+
+    .. versionchanged:: 2021.2
+
+        Can memoize methods on classes that do not allow setting attributes
+        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
 
-    return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
+    return memoize_on_first_arg(method,
+            intern(f"_memoize_dic_{method.__name__}"))
 
 
 class keyed_memoize_on_first_arg:  # noqa: N801
@@ -714,6 +732,8 @@ class keyed_memoize_on_first_arg:  # noqa: N801
 
     :arg key: A function receiving the same arguments as the decorated function
         which computes and returns the cache key.
+    :arg cache_dict_name: The name of the `dict` attribute in the instance
+        used to hold the cache.
 
     .. versionadded :: 2020.3
     """
@@ -723,8 +743,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
         self.cache_dict_name = cache_dict_name
 
     def _default_cache_dict_name(self, function):
-        return intern("_memoize_dic_"
-                + function.__module__ + function.__name__)
+        return intern(f"_memoize_dic_{function.__module__}{function.__name__}")
 
     def __call__(self, function):
         cache_dict_name = self.cache_dict_name
@@ -740,7 +759,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
                 return getattr(obj, cache_dict_name)[cache_key]
             except AttributeError:
                 result = function(obj, *args, **kwargs)
-                setattr(obj, cache_dict_name, {cache_key: result})
+                object.__setattr__(obj, cache_dict_name, {cache_key: result})
                 return result
             except KeyError:
                 result = function(obj, *args, **kwargs)
@@ -748,7 +767,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
                 return result
 
         def clear_cache(obj):
-            delattr(obj, cache_dict_name)
+            object.__delattr__(obj, cache_dict_name)
 
         from functools import update_wrapper
         new_wrapper = update_wrapper(wrapper, function)
@@ -758,15 +777,23 @@ class keyed_memoize_on_first_arg:  # noqa: N801
 
 
 class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
-    """Supports cache deletion via ``method_name.clear_cache(self)``.
+    """Like :class:`memoize_method`, but additionally uses a function *key* to
+    compute the key under which the function result is stored.
+
+    Supports cache deletion via ``method_name.clear_cache(self)``.
 
     :arg key: A function receiving the same arguments as the decorated function
         which computes and returns the cache key.
 
     .. versionadded :: 2020.3
+
+    .. versionchanged:: 2021.2
+
+        Can memoize methods on classes that do not allow setting attributes
+        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
     def _default_cache_dict_name(self, function):
-        return intern("_memoize_dic_" + function.__name__)
+        return intern(f"_memoize_dic_{function.__name__}")
 
 
 def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
@@ -791,7 +818,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     uncached_kwargs = list(uncached_kwargs)
 
     def parametrized_decorator(method):
-        cache_dict_name = intern("_memoize_dic_"+method.__name__)
+        cache_dict_name = intern(f"_memoize_dic_{method.__name__}")
 
         def wrapper(self, *args, **kwargs):
             cache_args = list(args)
@@ -817,7 +844,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
                 return getattr(self, cache_dict_name)[key]
             except AttributeError:
                 result = method(self, *args, **kwargs)
-                setattr(self, cache_dict_name, {key: result})
+                object.__setattr__(self, cache_dict_name, {key: result})
                 return result
             except KeyError:
                 result = method(self, *args, **kwargs)
@@ -825,7 +852,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
                 return result
 
         def clear_cache(self):
-            delattr(self, cache_dict_name)
+            object.__delattr__(self, cache_dict_name)
 
         if sys.version_info >= (2, 5):
             from functools import update_wrapper
@@ -837,55 +864,28 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     return parametrized_decorator
 
 
-def memoize_method_nested(inner):
-    """Adds a cache to a function nested inside a method. The cache is attached
-    to *memoize_cache_context* (if it exists) or *self* in the outer (method)
-    namespace.
-
-    Requires Python 2.5 or newer.
-    """
-
-    from warnings import warn
-    warn("memoize_method_nested is deprecated and will go away in 2021. "
-            "Use @memoize_in(self, 'identifier') instead", DeprecationWarning,
-            stacklevel=2)
-
-    cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
-            % (inner.__name__, inner.__code__.co_filename,
-                inner.__code__.co_firstlineno))
-
-    from inspect import currentframe
-    outer_frame = currentframe().f_back
-    cache_context = outer_frame.f_locals.get("memoize_cache_context")
-    if cache_context is None:
-        cache_context = outer_frame.f_locals.get("self")
-
-    try:
-        cache_dict = getattr(cache_context, cache_dict_name)
-    except AttributeError:
-        cache_dict = {}
-        setattr(cache_context, cache_dict_name, cache_dict)
-
-    @wraps(inner)
-    def new_inner(*args):
-        try:
-            return cache_dict[args]
-        except KeyError:
-            result = inner(*args)
-            cache_dict[args] = result
-            return result
-
-    return new_inner
-
-
 class memoize_in:  # noqa
-    """Adds a cache to a function nested inside a method. The cache is attached
-    to *container*.
+    """Adds a cache to the function it decorates. The cache is attached
+    to *container* and must be uniquely specified by *identifier* (i.e.
+    all functions using the same *container* and *identifier* will be using
+    the same cache). The decorated function may only receive positional
+    arguments.
+
+    .. note::
+
+        This function works well on nested functions, which
+        do not have stable global identifiers.
 
     .. versionchanged :: 2020.3
 
         *identifier* no longer needs to be a :class:`str`,
         but it needs to be hashable.
+
+    .. versionchanged:: 2021.2.1
+
+        Can now use instances of classes as *container* that do not allow
+        setting attributes (e.g. by overwritting ``__setattr__``),
+        e.g. frozen :mod:`dataclasses`.
     """
 
     def __init__(self, container, identifier):
@@ -893,7 +893,8 @@ class memoize_in:  # noqa
             memoize_in_dict = container._pytools_memoize_in_dict
         except AttributeError:
             memoize_in_dict = {}
-            container._pytools_memoize_in_dict = memoize_in_dict
+            object.__setattr__(container, "_pytools_memoize_in_dict",
+                    memoize_in_dict)
 
         self.cache_dict = memoize_in_dict.setdefault(identifier, {})
 
@@ -909,6 +910,41 @@ class memoize_in:  # noqa
 
         return new_inner
 
+
+class keyed_memoize_in:  # noqa
+    """Like :class:`memoize_in`, but additionally uses a function *key* to
+    compute the key under which the function result is memoized.
+
+    :arg key: A function receiving the same arguments as the decorated function
+        which computes and returns the cache key.
+
+    .. versionadded :: 2021.2.1
+    """
+
+    def __init__(self, container, identifier, key):
+        try:
+            memoize_in_dict = container._pytools_keyed_memoize_in_dict
+        except AttributeError:
+            memoize_in_dict = {}
+            object.__setattr__(container, "_pytools_keyed_memoize_in_dict",
+                    memoize_in_dict)
+
+        self.cache_dict = memoize_in_dict.setdefault(identifier, {})
+        self.key = key
+
+    def __call__(self, inner):
+        @wraps(inner)
+        def new_inner(*args):
+            key = self.key(*args)
+            try:
+                return self.cache_dict[key]
+            except KeyError:
+                result = inner(*args)
+                self.cache_dict[key] = result
+                return result
+
+        return new_inner
+
 # }}}
 
 
@@ -1394,22 +1430,8 @@ def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
                 yield remainder + (i,)
 
 
-def generate_all_nonnegative_integer_tuples(length, least=0):
-    assert length >= 0
-    current_max = least
-    while True:
-        for max_pos in range(length):
-            for prebase in generate_nonnegative_integer_tuples_below(
-                    current_max, max_pos, least):
-                for postbase in generate_nonnegative_integer_tuples_below(
-                        current_max+1, length-max_pos-1, least):
-                    yield prebase + [current_max] + postbase
-        current_max += 1
-
-
 # backwards compatibility
 generate_positive_integer_tuples_below = generate_nonnegative_integer_tuples_below
-generate_all_positive_integer_tuples = generate_all_nonnegative_integer_tuples
 
 
 def _pos_and_neg_adaptor(tuple_iter):
@@ -1429,11 +1451,6 @@ def generate_all_integer_tuples_below(n, length, least_abs=0):
         n, length, least_abs))
 
 
-def generate_all_integer_tuples(length, least_abs=0):
-    return _pos_and_neg_adaptor(generate_all_nonnegative_integer_tuples(
-        length, least_abs))
-
-
 def generate_permutations(original):
     """Generate all permutations of the list *original*.
 
@@ -2087,6 +2104,9 @@ def generate_unique_names(prefix):
         try_num += 1
 
 
+UNIQUE_NAME_GEN_COUNTER_RE = re.compile(r"^(?P<based_on>\w+)_(?P<counter>\d+)$")
+
+
 def generate_numbered_unique_names(
         prefix: str, num: Optional[int] = None) -> Iterable[Tuple[int, str]]:
     if num is None:
@@ -2135,9 +2155,10 @@ class UniqueNameGenerator:
 
     def add_name(self, name: str) -> None:
         if self.is_name_conflicting(name):
-            raise ValueError("name '%s' conflicts with existing names")
+            raise ValueError("name '%s' conflicts with existing names" % name)
         if not name.startswith(self.forced_prefix):
-            raise ValueError("name '%s' does not start with required prefix")
+            raise ValueError("name '%s' does not start with required prefix '%s'"
+                             % (name, self.forced_prefix))
 
         self.existing_names.add(name)
         self._name_added(name)
@@ -2151,7 +2172,18 @@ class UniqueNameGenerator:
 
         counter = self.prefix_to_counter.get(based_on, None)
 
-        for counter, var_name in generate_numbered_unique_names(based_on, counter):
+        # {{{ try to get counter from based_on if not already present
+
+        if counter is None:
+            counter_match = UNIQUE_NAME_GEN_COUNTER_RE.match(based_on)
+
+            if counter_match:
+                based_on = counter_match.groupdict()["based_on"]
+                counter = int(counter_match.groupdict()["counter"])
+
+        # }}}
+
+        for counter, var_name in generate_numbered_unique_names(based_on, counter):  # noqa: B007,E501
             if not self.is_name_conflicting(var_name):
                 break
 
@@ -2299,7 +2331,7 @@ def reshaped_view(a, newshape):
 
 # {{{ process timer
 
-SUPPORTS_PROCESS_TIME = (sys.version_info >= (3, 3))
+SUPPORTS_PROCESS_TIME = True
 
 
 class ProcessTimer:
@@ -2314,20 +2346,16 @@ class ProcessTimer:
     .. attribute:: wall_elapsed
     .. attribute:: process_elapsed
 
-        Only available in Python 3.3+.
-
     .. versionadded:: 2018.5
     """
 
     def __init__(self):
         import time
-        if SUPPORTS_PROCESS_TIME:
-            self.perf_counter_start = time.perf_counter()
-            self.process_time_start = time.process_time()
+        self.perf_counter_start = time.perf_counter()
+        self.process_time_start = time.process_time()
 
-        else:
-            import timeit
-            self.time_start = timeit.default_timer()
+        self.wall_elapsed = None
+        self.process_elapsed = None
 
     def __enter__(self):
         return self
@@ -2336,17 +2364,20 @@ class ProcessTimer:
         self.done()
 
     def done(self):
-        # pylint: disable=attribute-defined-outside-init
-
         import time
-        if SUPPORTS_PROCESS_TIME:
-            self.wall_elapsed = time.perf_counter() - self.perf_counter_start
-            self.process_elapsed = time.process_time() - self.process_time_start
+        self.wall_elapsed = time.perf_counter() - self.perf_counter_start
+        self.process_elapsed = time.process_time() - self.process_time_start
 
-        else:
-            import timeit
-            self.wall_elapsed = timeit.default_timer() - self.time_start
-            self.process_elapsed = None
+    def __str__(self):
+        cpu = self.process_elapsed / self.wall_elapsed
+        return f"{self.wall_elapsed:.2f}s wall {cpu:.2f}x CPU"
+
+    def __repr__(self):
+        wall = self.wall_elapsed
+        process = self.process_elapsed
+
+        return (f"{type(self).__name__}"
+                f"(wall_elapsed={wall!r}s, process_elapsed={process!r}s)")
 
 # }}}
 
@@ -2397,7 +2428,11 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
             # Can happen, e.g., if pudb is controlling the console.
             use_late_start_logging = False
         else:
-            use_late_start_logging = sys.stdin.isatty()
+            if hasattr(sys.stdin, "closed") and not sys.stdin.closed:
+                # can query stdin.isatty() only if stdin's open
+                use_late_start_logging = sys.stdin.isatty()
+            else:
+                use_late_start_logging = False
 
         import os
         if os.environ.get("PYTOOLS_LOG_NO_THREADS", ""):
@@ -2434,20 +2469,13 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
         self.timer.done()
         self.is_done = True
 
-        wall_elapsed = self.timer.wall_elapsed
-        process_elapsed = self.timer.process_elapsed
-
         completion_level = (
                 self.noisy_level
-                if wall_elapsed > self.long_threshold_seconds
+                if self.timer.wall_elapsed > self.long_threshold_seconds
                 else self.silent_level)
 
-        if process_elapsed is not None:
-            msg = "%s: completed (%.2fs wall, %.1fx CPU)"
-            fmt_args = [self.description, wall_elapsed, process_elapsed/wall_elapsed]
-        else:
-            msg = "%s: completed (%f.2s wall)"
-            fmt_args = [self.description, wall_elapsed]
+        msg = "%s: completed (%s)"
+        fmt_args = [self.description, str(self.timer)]
 
         if extra_msg:
             msg += ": " + extra_msg
@@ -2543,6 +2571,107 @@ def natsorted(iterable, key=None, reverse=False):
 # }}}
 
 
+# {{{ resolve_name
+
+# https://github.com/python/cpython/commit/1ed61617a4a6632905ad6a0b440cd2cafb8b6414
+
+_DOTTED_WORDS = r"[a-z_]\w*(\.[a-z_]\w*)*"
+_NAME_PATTERN = re.compile(f"^({_DOTTED_WORDS})(:({_DOTTED_WORDS})?)?$", re.I)
+del _DOTTED_WORDS
+
+
+def resolve_name(name):
+    """A backport of :func:`pkgutil.resolve_name` (added in Python 3.9).
+
+    .. versionadded:: 2021.1.2
+    """
+    # Delete the tail of the function and deprecate this once we require Python 3.9.
+    if sys.version_info >= (3, 9):
+        # use the official version
+        import pkgutil
+        return pkgutil.resolve_name(name)  # pylint: disable=no-member
+
+    import importlib
+
+    m = _NAME_PATTERN.match(name)
+    if not m:
+        raise ValueError(f"invalid format: {name!r}")
+    groups = m.groups()
+    if groups[2]:
+        # there is a colon - a one-step import is all that's needed
+        mod = importlib.import_module(groups[0])
+        parts = groups[3].split(".") if groups[3] else []
+    else:
+        # no colon - have to iterate to find the package boundary
+        parts = name.split(".")
+        modname = parts.pop(0)
+        # first part *must* be a module/package.
+        mod = importlib.import_module(modname)
+        while parts:
+            p = parts[0]
+            s = f"{modname}.{p}"
+            try:
+                mod = importlib.import_module(s)
+                parts.pop(0)
+                modname = s
+            except ImportError:
+                break
+    # if we reach this point, mod is the module, already imported, and
+    # parts is the list of parts in the object hierarchy to be traversed, or
+    # an empty list if just the module is wanted.
+    result = mod
+    for p in parts:
+        result = getattr(result, p)
+    return result
+
+# }}}
+
+
+# {{{ unordered_hash
+
+def unordered_hash(hash_instance, iterable, hash_constructor=None):
+    """Using a hash algorithm given by the parameter-less constructor
+    *hash_constructor*, return a hash object whose internal state
+    depends on the entries of *iterable*, but not their order. If *hash*
+    is the instance returned by evaluating ``hash_constructor()``, then
+    the each entry *i* of the iterable must permit ``hash.upate(i)`` to
+    succeed. An example of *hash_constructor* is ``hashlib.sha256``
+    from :mod:`hashlib`.  ``hash.digest_size`` must also be defined.
+    If *hash_constructor* is not provided, ``hash_instance.name`` is
+    used to deduce it.
+
+    :returns: the updated *hash_instance*.
+
+    .. warning::
+
+        The construction used in this function is likely not cryptographically
+        secure. Do not use this function in a security-relevant context.
+
+    .. versionadded:: 2021.2
+    """
+
+    if hash_constructor is None:
+        from functools import partial
+        import hashlib
+        hash_constructor = partial(hashlib.new, hash_instance.name)
+
+    h_int = 0
+    for i in iterable:
+        h_i = hash_constructor()
+        h_i.update(i)
+        # Using sys.byteorder (for efficiency) here technically makes the
+        # hash system-dependent (which it should not be), however the
+        # effect of this is undone by the to_bytes conversion below, while
+        # left invariant by the intervening XOR operations (which do not
+        # mix adjacent bits).
+        h_int = h_int ^ int.from_bytes(h_i.digest(), sys.byteorder)
+
+    hash_instance.update(h_int.to_bytes(hash_instance.digest_size, sys.byteorder))
+    return hash_instance
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
